import os
from typing import Type

from asphalt.core import Component, Context

from jupyverse_api.app import App
from jupyverse_api.auth import Auth
from jupyverse_api.terminals import Terminals, TerminalServer

from .routes import _Terminals

_TerminalServer: Type[TerminalServer]
if os.name == "nt":
    from .win_server import _TerminalServer
else:
    from .server import _TerminalServer


class TerminalsComponent(Component):
    async def start(
        self,
        ctx: Context,
    ) -> None:
        app = await ctx.request_resource(App)
        auth = await ctx.request_resource(Auth)  # type: ignore

        terminals = _Terminals(app, auth, _TerminalServer)
        ctx.add_resource(terminals, types=Terminals)



from typing import Optional

from pydantic import BaseModel


class User(BaseModel):
    username: str = ""
    name: str = ""
    display_name: str = ""
    initials: Optional[str] = None
    color: Optional[str] = None
    avatar_url: Optional[str] = None
    workspace: str = "{}"
    settings: str = "{}"



from copy import deepcopy
from typing import Dict, List, Optional
from uuid import uuid4


async def authenticate_client(http, port, permissions={}):
    # create a new user
    username = uuid4().hex
    # if logged in, log out
    first_time = True
    while True:
        response = await http.get(f"http://127.0.0.1:{port}/api/me")
        if response.status_code == 403:
            break
        assert first_time
        response = await http.post(f"http://127.0.0.1:{port}/auth/logout")
        assert response.status_code == 200
        first_time = False

    # register user
    register_body = {
        "email": f"{username}@example.com",
        "password": username,
        "username": username,
        "permissions": permissions,
    }
    response = await http.post(f"http://127.0.0.1:{port}/auth/register", json=register_body)
    # check that we cannot register if not logged in
    assert response.status_code == 403
    # log in as admin
    login_body = {"username": "admin@jupyter.com", "password": "jupyverse"}
    response = await http.post(f"http://127.0.0.1:{port}/auth/login", data=login_body)
    assert response.status_code == 204
    # register user
    response = await http.post(f"http://127.0.0.1:{port}/auth/register", json=register_body)
    assert response.status_code == 201

    # log out
    response = await http.post(f"http://127.0.0.1:{port}/auth/logout")
    assert response.status_code == 204
    # check that we can't get our identity, since we're not logged in
    response = await http.get(f"http://127.0.0.1:{port}/api/me")
    assert response.status_code == 403

    # log in with registered user
    login_body = {"username": f"{username}@example.com", "password": username}
    response = await http.post(f"http://127.0.0.1:{port}/auth/login", data=login_body)
    assert response.status_code == 204
    # we should now have a cookie
    assert "fastapiusersauth" in http.cookies
    # check our identity, since we're logged in
    response = await http.get(
        f"http://127.0.0.1:{port}/api/me", params={"permissions": permissions}
    )
    assert response.status_code == 200
    me = response.json()
    assert me["identity"]["username"] == username
    # check our permissions
    assert me["permissions"] == permissions


def configure(components, config):
    # TODO: generalize to arbitrary nested dictionaries, not just one level
    _components = deepcopy(components)
    for k1, v1 in config.items():
        for k2, v2 in v1.items():
            _components[k1][k2] = v2
    return _components


def create_content(
    content: Optional[List],
    type: str,
    size: Optional[int],
    mimetype: Optional[str],
    name: str,
    path: str,
    format: Optional[str],
) -> Dict:
    return {
        "content": content,
        "created": None,
        "format": format,
        "last_modified": None,
        "mimetype": mimetype,
        "name": name,
        "path": path,
        "size": size,
        "type": type,
        "writable": True,
    }


def clear_content_values(content: Dict, keys: List[str] = []):
    for k in content:
        if k in keys:
            content[k] = None
        if k == "content" and isinstance(content[k], list):
            for c in content[k]:
                clear_content_values(c, keys)
    return content


def sort_content_by_name(content: Dict):
    for k in content:
        if k == "content" and isinstance(content[k], list):
            # FIXME: this sorting algorithm is terrible!
            names = [c["name"] for c in content[k]]
            names.sort()
            new_content = []
            for name in names:
                for i, c in enumerate(content[k]):
                    if c["name"] == name:
                        break
                content[k].pop(i)
                new_content.append(c)
            content[k] = new_content
            for c in content[k]:
                sort_content_by_name(c)
    return content



from pydantic import Field

from jupyverse_api.auth import AuthConfig


class AuthJupyterHubConfig(AuthConfig):
    db_url: str = Field(
        description="The connection URL passed to create_engine()",
        default="sqlite+aiosqlite:///:memory:",
    )



import json
from typing import Any, Callable, Dict, List, Optional, Tuple

import httpx
from fastapi import APIRouter, Depends, Query, Request, Response
from fastapi.responses import RedirectResponse
from fief_client import FiefAccessTokenInfo

from jupyverse_api import Router
from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User

from .backend import get_backend
from .config import _AuthFiefConfig


def auth_factory(
    app: App,
    auth_fief_config: _AuthFiefConfig,
):
    backend = get_backend(auth_fief_config)

    class _AuthFief(Auth, Router):
        def __init__(self) -> None:
            super().__init__(app)

            router = APIRouter()

            @router.get("/auth-callback", name="auth_callback")
            async def auth_callback(
                request: Request,
                response: Response,
                code: str = Query(...),
            ):
                if auth_fief_config.callback_url:
                    redirect_uri = auth_fief_config.callback_url
                else:
                    redirect_uri = str(request.url_for("auth_callback"))
                tokens, user_info = await backend.fief.auth_callback(code, redirect_uri)

                user_id = user_info["sub"]
                async with httpx.AsyncClient() as client:
                    headers = {"Authorization": f"Bearer {auth_fief_config.admin_api_key}"}
                    r = await client.get(
                        f"{auth_fief_config.base_url}/admin/api/oauth-providers/{auth_fief_config.oauth_provider_id}/access-token/{user_id}",
                        headers=headers,
                    )

                # FIXME: this is hard-coded for GitHub authentication
                access_token = r.json()["access_token"]
                async with httpx.AsyncClient() as client:
                    headers = {
                        "Authorization": f"Bearer {access_token}",
                        "Accept": "application/vnd.github+json",
                        "X-GitHub-Api-Version": "2022-11-28",
                    }
                    r = await client.get("https://api.github.com/user", headers=headers)

                d = r.json()
                data = {
                    "username": d["login"],
                    "name": d["name"],
                    "display_name": d["name"],
                    "initials": "".join([word[0] for word in d["name"].split()]),
                    "avatar_url": d["avatar_url"],
                }
                await backend.fief.update_profile(tokens["access_token"], {"fields": data})

                # set permissions
                async with httpx.AsyncClient() as client:
                    headers = {
                        "Authorization": f"Bearer {auth_fief_config.admin_api_key}",
                        "accept": "application/json",
                    }
                    skip = 0
                    nb = 100
                    count = None
                    got = 0
                    while True:
                        params = {"limit": nb, "skip": skip}
                        r = await client.get(
                            f"{auth_fief_config.base_url}/admin/api/permissions/",
                            headers=headers,
                            params=params,
                        )
                        d = r.json()
                        if count is None:
                            count = d["count"]

                        permission_id = {
                            result["codename"]: result["id"] for result in d["results"]
                        }

                        headers = {"Authorization": f"Bearer {auth_fief_config.admin_api_key}"}
                        for permission, id in permission_id.items():
                            data = {"id": id}
                            await client.post(
                                f"{auth_fief_config.base_url}/admin/api/users/{user_id}/permissions",
                                headers=headers,
                                json=data,
                            )

                        got += len(d["results"])
                        if got < count:
                            skip += nb
                        else:
                            break

                refresh_token = tokens["refresh_token"]
                assert refresh_token is not None
                tokens, user_info = await backend.fief.auth_refresh_token(refresh_token)

                response = RedirectResponse(request.url_for("root"))
                response.set_cookie(
                    backend.session_cookie_name,
                    tokens["access_token"],
                    max_age=tokens["expires_in"],
                    httponly=True,
                    secure=False,
                )

                return response

            @router.get("/api/me")
            async def get_api_me(
                request: Request,
                user: User = Depends(self.current_user()),
                access_token_info: FiefAccessTokenInfo = Depends(backend.auth.authenticated()),
            ):
                checked_permissions: Dict[str, List[str]] = {}
                permissions = json.loads(
                    dict(request.query_params).get("permissions", "{}").replace("'", '"')
                )
                if permissions:
                    user_permissions: Dict[str, List[str]] = {}
                    for permission in access_token_info["permissions"]:
                        resource, action = permission.split(":")
                        if resource not in user_permissions:
                            user_permissions[resource] = []
                        user_permissions[resource].append(action)
                    for resource, actions in permissions.items():
                        user_resource_permissions = user_permissions.get(resource, [])
                        allowed = checked_permissions[resource] = []
                        for action in actions:
                            if action in user_resource_permissions:
                                allowed.append(action)

                keys = ["username", "name", "display_name", "initials", "avatar_url", "color"]
                identity = {k: getattr(user, k) for k in keys}
                return {
                    "identity": identity,
                    "permissions": checked_permissions,
                }

            self.include_router(router)

        def current_user(self, permissions: Optional[Dict[str, List[str]]] = None) -> Callable:
            return backend.current_user(permissions)

        async def update_user(self, update_user=Depends(backend.update_user)) -> Callable:
            return update_user

        def websocket_auth(
            self,
            permissions: Optional[Dict[str, List[str]]] = None,
        ) -> Callable[[], Tuple[Any, Dict[str, List[str]]]]:
            return backend.websocket_auth(permissions)

    return _AuthFief()



from __future__ import annotations

from contextlib import AsyncExitStack
from functools import partial
from logging import Logger, getLogger

from anyio import (
    TASK_STATUS_IGNORED,
    Event,
    create_memory_object_stream,
    create_task_group,
)
from anyio.abc import TaskGroup, TaskStatus
from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream
from pycrdt import Doc

from .websocket import Websocket
from .yutils import (
    YMessageType,
    create_update_message,
    process_sync_message,
    put_updates,
    sync,
)


class WebsocketProvider:
    """WebSocket provider."""

    _ydoc: Doc
    _update_send_stream: MemoryObjectSendStream
    _update_receive_stream: MemoryObjectReceiveStream
    _started: Event | None
    _starting: bool
    _task_group: TaskGroup | None

    def __init__(self, ydoc: Doc, websocket: Websocket, log: Logger | None = None) -> None:
        """Initialize the object.

        The WebsocketProvider instance should preferably be used as an async context manager:
        ```py
        async with websocket_provider:
            ...
        ```
        However, a lower-level API can also be used:
        ```py
        task = asyncio.create_task(websocket_provider.start())
        await websocket_provider.started.wait()
        ...
        websocket_provider.stop()
        ```

        Arguments:
            ydoc: The YDoc to connect through the WebSocket.
            websocket: The WebSocket through which to connect the YDoc.
            log: An optional logger.
        """
        self._ydoc = ydoc
        self._websocket = websocket
        self.log = log or getLogger(__name__)
        self._update_send_stream, self._update_receive_stream = create_memory_object_stream(
            max_buffer_size=65536
        )
        self._started = None
        self._starting = False
        self._task_group = None
        ydoc.observe(partial(put_updates, self._update_send_stream))

    @property
    def started(self) -> Event:
        """An async event that is set when the WebSocket provider has started."""
        if self._started is None:
            self._started = Event()
        return self._started

    async def __aenter__(self) -> WebsocketProvider:
        if self._task_group is not None:
            raise RuntimeError("WebsocketProvider already running")

        async with AsyncExitStack() as exit_stack:
            tg = create_task_group()
            self._task_group = await exit_stack.enter_async_context(tg)
            self._exit_stack = exit_stack.pop_all()
            tg.start_soon(self._run)
            self.started.set()

        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        if self._task_group is None:
            raise RuntimeError("WebsocketProvider not running")

        self._task_group.cancel_scope.cancel()
        self._task_group = None
        return await self._exit_stack.__aexit__(exc_type, exc_value, exc_tb)

    async def _run(self):
        await sync(self._ydoc, self._websocket, self.log)
        self._task_group.start_soon(self._send)
        async for message in self._websocket:
            if message[0] == YMessageType.SYNC:
                await process_sync_message(message[1:], self._ydoc, self._websocket, self.log)

    async def _send(self):
        async with self._update_receive_stream:
            async for update in self._update_receive_stream:
                message = create_update_message(update)
                try:
                    await self._websocket.send(message)
                except Exception:
                    pass

    async def start(self, *, task_status: TaskStatus[None] = TASK_STATUS_IGNORED):
        """Start the WebSocket provider.

        Arguments:
            task_status: The status to set when the task has started.
        """
        if self._starting:
            return
        else:
            self._starting = True

        if self._task_group is not None:
            raise RuntimeError("WebsocketProvider already running")

        async with create_task_group() as self._task_group:
            self._task_group.start_soon(self._run)
            self.started.set()
            self._starting = False
            task_status.started()

    def stop(self):
        """Stop the WebSocket provider."""
        if self._task_group is None:
            raise RuntimeError("WebsocketProvider not running")

        self._task_group.cancel_scope.cancel()
        self._task_group = None



from __future__ import annotations

import asyncio
import logging
from datetime import datetime
from functools import partial
from typing import Dict
from uuid import uuid4

from fastapi import (
    HTTPException,
    Request,
    Response,
    WebSocketDisconnect,
    status,
)
from pycrdt import Doc
from websockets.exceptions import ConnectionClosedOK

from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User
from jupyverse_api.contents import Contents
from jupyverse_api.yjs import Yjs, YjsConfig
from jupyverse_api.yjs.models import CreateDocumentSession

from .ydocs import ydocs as YDOCS
from .ydocs.ybasedoc import YBaseDoc
from .ywebsocket.websocket_server import WebsocketServer, YRoom
from .ywebsocket.ystore import SQLiteYStore, YDocNotFound
from .ywebsocket.yutils import YMessageType, YSyncMessageType
from .ywidgets import Widgets

YFILE = YDOCS["file"]
AWARENESS = 1
SERVER_SESSION = uuid4().hex
logger = logging.getLogger("yjs")


class JupyterSQLiteYStore(SQLiteYStore):
    db_path = ".jupyter_ystore.db"  # FIXME: pass in config


class _Yjs(Yjs):
    def __init__(
        self,
        app: App,
        yjs_config: YjsConfig,
        auth: Auth,
        contents: Contents,
    ) -> None:
        super().__init__(app=app, auth=auth)
        self.contents = contents
        self.room_manager = RoomManager(contents, yjs_config)
        if Widgets is None:
            self.widgets = None
        else:
            self.widgets = Widgets()  # type: ignore

    async def collaboration_room_websocket(
        self,
        path,
        websocket_permissions,
    ):
        if websocket_permissions is None:
            return
        websocket, permissions = websocket_permissions
        await websocket.accept()
        ywebsocket = YWebsocket(websocket, path)
        await self.room_manager.serve(ywebsocket, permissions)

    async def create_roomid(
        self,
        path,
        request: Request,
        response: Response,
        user: User,
    ):
        # we need to process the request manually
        # see https://github.com/tiangolo/fastapi/issues/3373#issuecomment-1306003451
        create_document_session = CreateDocumentSession(**(await request.json()))
        idx = await self.contents.file_id_manager.get_id(path)
        res = {
            "format": create_document_session.format,
            "type": create_document_session.type,
            "sessionId": SERVER_SESSION,
        }
        if idx is not None:
            res["fileId"] = idx
            return res

        idx = await self.contents.file_id_manager.index(path)
        if idx is None:
            raise HTTPException(status_code=404, detail=f"File {path} does not exist")

        response.status_code = status.HTTP_201_CREATED
        res["fileId"] = idx
        return res

    def get_document(self, document_id: str) -> YBaseDoc:
        return self.room_manager.documents[document_id]


def to_datetime(iso_date: str) -> datetime:
    return datetime.fromisoformat(iso_date.rstrip("Z"))


class YWebsocket:
    """An wrapper to make a Starlette's WebSocket look like a ywebsocket's WebSocket"""

    def __init__(self, websocket, path: str):
        self._websocket = websocket
        self._path = path

    @property
    def path(self) -> str:
        return self._path

    @path.setter
    def path(self, value: str) -> None:
        self._path = value

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            message = await self._websocket.receive_bytes()
        except WebSocketDisconnect:
            raise StopAsyncIteration()
        return message

    async def send(self, message):
        try:
            await self._websocket.send_bytes(message)
        except ConnectionClosedOK:
            return

    async def recv(self):
        return await self._websocket.receive_bytes()


class RoomManager:
    contents: Contents
    documents: Dict[str, YBaseDoc]
    watchers: Dict[str, asyncio.Task]
    savers: Dict[str, asyncio.Task]
    cleaners: Dict[YRoom, asyncio.Task]
    last_modified: Dict[str, datetime]
    websocket_server: JupyterWebsocketServer
    lock: asyncio.Lock

    def __init__(self, contents: Contents, yjs_config: YjsConfig):
        self.contents = contents
        self.yjs_config = yjs_config
        self.documents = {}  # a dictionary of room_name:document
        self.watchers = {}  # a dictionary of file_id:task
        self.savers = {}  # a dictionary of file_id:task
        self.cleaners = {}  # a dictionary of room:task
        self.last_modified = {}  # a dictionary of file_id:last_modification_date
        self.websocket_server = JupyterWebsocketServer(rooms_ready=False, auto_clean_rooms=False)
        self.websocket_server_task = asyncio.create_task(self.websocket_server.start())
        self.lock = asyncio.Lock()

    def stop(self):
        for watcher in self.watchers.values():
            watcher.cancel()
        for saver in self.savers.values():
            saver.cancel()
        for cleaner in self.cleaners.values():
            cleaner.cancel()
        self.websocket_server.stop()

    async def serve(self, websocket: YWebsocket, permissions) -> None:
        room = await self.websocket_server.get_room(websocket.path)
        can_write = permissions is None or "write" in permissions.get("yjs", [])
        room.on_message = partial(self.filter_message, can_write)
        is_stored_document = websocket.path.count(":") >= 2
        if is_stored_document:
            assert room.ystore is not None
            file_format, file_type, file_id = websocket.path.split(":", 2)
            if room in self.cleaners:
                # cleaning the room was scheduled because there was no client left
                # cancel that since there is a new client
                self.cleaners[room].cancel()
            if not room.ready:
                file_path = await self.contents.file_id_manager.get_path(file_id)
                logger.info(f"Opening collaboration room: {websocket.path} ({file_path})")
                document = YDOCS.get(file_type, YFILE)(room.ydoc)
                document.file_id = file_id
                self.documents[websocket.path] = document
                async with self.lock:
                    model = await self.contents.read_content(file_path, True, file_format)
                assert model.last_modified is not None
                self.last_modified[file_id] = to_datetime(model.last_modified)
                if not room.ready:
                    # try to apply Y updates from the YStore for this document
                    try:
                        await room.ystore.apply_updates(room.ydoc)
                        read_from_source = False
                    except YDocNotFound:
                        # YDoc not found in the YStore, create the document from
                        # the source file (no change history)
                        read_from_source = True
                    if not read_from_source:
                        # if YStore updates and source file are out-of-sync, resync updates
                        # with source
                        if document.source != model.content:
                            read_from_source = True
                    if read_from_source:
                        document.source = model.content
                        await room.ystore.encode_state_as_update(room.ydoc)

                    document.dirty = False
                    room.ready = True
                    # save the document to file when changed
                    document.observe(
                        partial(self.on_document_change, file_id, file_type, file_format, document)
                    )
                    # update the document when file changes
                    if file_id not in self.watchers:
                        self.watchers[file_id] = asyncio.create_task(
                            self.watch_file(file_format, file_id, document)
                        )

        await self.websocket_server.started.wait()
        await self.websocket_server.serve(websocket)

        if is_stored_document and not room.clients:
            # no client in this room after we disconnect
            self.cleaners[room] = asyncio.create_task(self.maybe_clean_room(room, websocket.path))

    async def filter_message(self, can_write: bool, message: bytes) -> bool:
        """
        Called whenever a message is received, before forwarding it to other clients.

        :param can_write: True if updating the document is permitted, False otherwise.
        :param message: received message.
        :returns: True if the message must be discarded, False otherwise (default: False).
        """
        skip = False
        byte = message[0]
        msg = message[1:]
        if byte == AWARENESS:
            # changes = self.room.awareness.get_changes(msg)
            # # filter out message depending on changes
            # skip = True
            pass
        elif byte == YMessageType.SYNC:
            if not can_write and msg[0] == YSyncMessageType.SYNC_UPDATE:
                skip = True
        else:
            skip = True
        return skip

    async def get_file_path(self, file_id: str, document) -> str | None:
        file_path = await self.contents.file_id_manager.get_path(file_id)
        if file_path is None:
            return
        if file_path != document.path:
            document.path = file_path
        return file_path

    async def watch_file(self, file_format: str, file_id: str, document: YBaseDoc) -> None:
        file_path = await self.get_file_path(file_id, document)
        assert file_path is not None
        logger.debug(f"Watching file: {file_path}")
        while True:
            watcher = self.contents.file_id_manager.watch(file_path)
            async for changes in watcher:
                new_file_path = await self.get_file_path(file_id, document)
                if new_file_path is None:
                    continue
                if new_file_path != file_path:
                    # file was renamed
                    self.contents.file_id_manager.unwatch(file_path, watcher)
                    file_path = new_file_path
                    # break
                await self.maybe_load_file(file_format, file_path, file_id)

    async def maybe_load_file(self, file_format: str, file_path: str, file_id: str) -> None:
        async with self.lock:
            model = await self.contents.read_content(file_path, False)
        # do nothing if the file was saved by us
        assert model.last_modified is not None
        if self.last_modified[file_id] < to_datetime(model.last_modified):
            # the file was not saved by us, update the shared document(s)
            async with self.lock:
                model = await self.contents.read_content(file_path, True, file_format)
            assert model.last_modified is not None
            documents = [v for k, v in self.documents.items() if k.split(":", 2)[2] == file_id]
            for document in documents:
                document.source = model.content
            self.last_modified[file_id] = to_datetime(model.last_modified)

    def on_document_change(
        self, file_id: str, file_type: str, file_format: str, document: YBaseDoc, target, event
    ) -> None:
        if target == "state" and "dirty" in event.keys:
            dirty = event.keys["dirty"]["newValue"]
            if not dirty:
                # we cleared the dirty flag, nothing to save
                return
        # unobserve and observe again because the structure of the document may have changed
        # e.g. a new cell added to a notebook
        document.unobserve()
        document.observe(
            partial(self.on_document_change, file_id, file_type, file_format, document)
        )
        if file_id in self.savers:
            self.savers[file_id].cancel()
        self.savers[file_id] = asyncio.create_task(
            self.maybe_save_document(file_id, file_type, file_format, document)
        )

    async def maybe_save_document(
        self, file_id: str, file_type: str, file_format: str, document: YBaseDoc
    ) -> None:
        # save after 1 second of inactivity to prevent too frequent saving
        await asyncio.sleep(self.yjs_config.document_save_delay)
        # if the room cannot be found, don't save
        try:
            file_path = await self.get_file_path(file_id, document)
        except Exception:
            return
        assert file_path is not None
        async with self.lock:
            model = await self.contents.read_content(file_path, True, file_format)
        assert model.last_modified is not None
        if self.last_modified[file_id] < to_datetime(model.last_modified):
            # file changed on disk, let's revert
            document.source = model.content
            self.last_modified[file_id] = to_datetime(model.last_modified)
            return
        if model.content != document.source:
            # don't save if not needed
            # this also prevents the dirty flag from bouncing between windows of
            # the same document opened as different types (e.g. notebook/text editor)
            content = {
                "content": document.source,
                "format": file_format,
                "path": file_path,
                "type": file_type,
            }
            async with self.lock:
                await self.contents.write_content(content)
                model = await self.contents.read_content(file_path, False)
            assert model.last_modified is not None
            self.last_modified[file_id] = to_datetime(model.last_modified)
        document.dirty = False
        # we're done saving, remove the saver
        del self.savers[file_id]

    async def maybe_clean_room(self, room, ws_path: str) -> None:
        # keep the document for a while in case someone reconnects
        await asyncio.sleep(self.yjs_config.document_cleanup_delay)
        document = self.documents[ws_path]
        document.unobserve()
        del self.documents[ws_path]
        file_id = ws_path.split(":", 2)[2]
        documents = [v for k, v in self.documents.items() if k.split(":", 2)[2] == file_id]
        if not documents:
            self.watchers[file_id].cancel()
            del self.watchers[file_id]
        room_name = self.websocket_server.get_room_name(room)
        self.websocket_server.delete_room(room=room)
        file_path = await self.get_file_path(file_id, document)
        logger.info(f"Closing collaboration room: {room_name} ({file_path})")


class JupyterWebsocketServer(WebsocketServer):
    async def get_room(self, ws_path: str, ydoc: Doc | None = None) -> YRoom:
        if ws_path not in self.rooms:
            if ws_path.count(":") >= 2:
                # it is a stored document (e.g. a notebook)
                file_format, file_type, file_id = ws_path.split(":", 2)
                updates_file_path = f".{file_type}:{file_id}.y"
                ystore = JupyterSQLiteYStore(path=updates_file_path)  # FIXME: pass in config
                self.rooms[ws_path] = YRoom(ydoc=ydoc, ready=False, ystore=ystore)
            else:
                # it is a transient document (e.g. awareness)
                self.rooms[ws_path] = YRoom(ydoc=ydoc)
        room = self.rooms[ws_path]
        await self.start_room(room)
        return room



from abc import ABC, abstractmethod
from typing import Any

from fastapi import APIRouter, Depends, Request, Response

from jupyverse_api import Config, Router

from ..app import App
from ..auth import Auth, User


class Yjs(Router, ABC):
    websocket_server: Any

    def __init__(self, app: App, auth: Auth):
        super().__init__(app=app)

        router = APIRouter()

        @router.websocket("/api/collaboration/room/{path:path}")
        async def collaboration_room_websocket(
            path,
            websocket_permissions=Depends(
                auth.websocket_auth(permissions={"yjs": ["read", "write"]})
            ),
        ):
            return await self.collaboration_room_websocket(path, websocket_permissions)

        @router.put("/api/collaboration/session/{path:path}", status_code=200)
        async def create_roomid(
            path,
            request: Request,
            response: Response,
            user: User = Depends(auth.current_user(permissions={"contents": ["read"]})),
        ):
            return await self.create_roomid(path, request, response, user)

        self.include_router(router)

    @abstractmethod
    async def collaboration_room_websocket(
        self,
        path,
        websocket_permissions,
    ):
        ...

    @abstractmethod
    async def create_roomid(
        self,
        path,
        request: Request,
        response: Response,
        user: User,
    ):
        ...

    @abstractmethod
    def get_document(
        self,
        document_id: str,
    ):
        ...


class YjsConfig(Config):
    document_cleanup_delay: float = 60
    document_save_delay: float = 1



import sys

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

ydocs = {ep.name: ep.load() for ep in entry_points(group="jupyverse_ydoc")}



from typing import Dict, List

from pydantic import ConfigDict

from jupyverse_api.auth import User


class JupyterHubUser(User):
    model_config = ConfigDict(from_attributes=True)

    token: str
    anonymous: bool = True
    permissions: Dict[str, List[str]]



import base64
from functools import partial
from typing import Any, Callable, Optional, Union

from pycrdt import Doc, Map

from .ybasedoc import YBaseDoc


class YBlob(YBaseDoc):
    """
    Extends :class:`YBaseDoc`, and represents a blob document.
    It is currently encoded as base64 because of:
    https://github.com/y-crdt/ypy/issues/108#issuecomment-1377055465
    The Y document can be set from bytes or from str, in which case it is assumed to be encoded as
    base64.
    """

    def __init__(self, ydoc: Optional[Doc] = None):
        super().__init__(ydoc)
        self._ysource = Map()
        self._ydoc["source"] = self._ysource

    @property
    def version(self) -> str:
        return "1.0.0"

    def get(self) -> bytes:
        return base64.b64decode(self._ysource["base64"].encode())

    def set(self, value: Union[bytes, str]) -> None:
        if isinstance(value, bytes):
            value = base64.b64encode(value).decode()
        self._ysource["base64"] = value

    def observe(self, callback: Callable[[str, Any], None]) -> None:
        self.unobserve()
        self._subscriptions[self._ystate] = self._ystate.observe(partial(callback, "state"))
        self._subscriptions[self._ysource] = self._ysource.observe(partial(callback, "source"))



import contextlib
import json
import logging
import random
from typing import Any, Callable, Dict, List, Optional, Tuple

from fastapi import APIRouter, Depends
from sqlalchemy import select  # type: ignore

from jupyverse_api import Router
from jupyverse_api.app import App
from jupyverse_api.auth import Auth
from jupyverse_api.frontend import FrontendConfig

from .backends import get_backend
from .config import _AuthConfig
from .db import get_db
from .models import UserCreate, UserRead, UserUpdate

logger = logging.getLogger("auth")


def auth_factory(
    app: App,
    auth_config: _AuthConfig,
    frontend_config: FrontendConfig,
):
    db = get_db(auth_config)
    backend = get_backend(auth_config, frontend_config, db)

    get_async_session_context = contextlib.asynccontextmanager(db.get_async_session)
    get_user_db_context = contextlib.asynccontextmanager(db.get_user_db)
    get_user_manager_context = contextlib.asynccontextmanager(backend.get_user_manager)

    @contextlib.asynccontextmanager
    async def _get_user_manager():
        async with get_async_session_context() as session:
            async with get_user_db_context(session) as user_db:
                async with get_user_manager_context(user_db) as user_manager:
                    yield user_manager

    async def create_user(**kwargs):
        async with _get_user_manager() as user_manager:
            await user_manager.create(UserCreate(**kwargs))

    async def update_user(user, **kwargs):
        async with _get_user_manager() as user_manager:
            await user_manager.update(UserUpdate(**kwargs), user)

    async def get_user_by_email(user_email):
        async with _get_user_manager() as user_manager:
            return await user_manager.get_by_email(user_email)

    class _Auth(Auth, Router):
        def __init__(self) -> None:
            super().__init__(app)

            self.db = db

            router = APIRouter()

            @router.get("/auth/users")
            async def get_users(
                user: UserRead = Depends(backend.current_user(permissions={"admin": ["read"]})),
            ):
                async with db.async_session_maker() as session:
                    statement = select(db.User)
                    users = (await session.execute(statement)).unique().all()
                return [usr.User for usr in users if usr.User.is_active]

            @router.get("/api/me")
            async def get_api_me(
                permissions: Optional[str] = None,
                user: UserRead = Depends(backend.current_user()),
                update_user = Depends(backend.update_user),
            ):
                checked_permissions: Dict[str, List[str]] = {}
                if permissions is None:
                    permissions = "{}"
                else:
                    permissions = permissions.replace("'", '"')
                permissions_dict = json.loads(permissions)
                if permissions_dict:
                    user_permissions = user.permissions
                    for resource, actions in permissions_dict.items():
                        user_resource_permissions = user_permissions.get(resource)
                        if user_resource_permissions is None:
                            continue
                        allowed = checked_permissions[resource] = []
                        for action in actions:
                            if action in user_resource_permissions:
                                allowed.append(action)

                keys = ["username", "name", "display_name", "initials", "avatar_url", "color"]
                identity = {k: getattr(user, k) for k in keys}
                if not identity["name"] and not identity["display_name"]:
                    moon = get_anonymous_username()
                    identity["name"] = f"Anonymous {moon}"
                    identity["display_name"] = f"Anonymous {moon}"
                    identity["initials"] = f"A{moon[0]}"
                    await update_user(
                        dict(
                            name=identity["name"],
                            display_name=identity["display_name"],
                            permissions=checked_permissions,
                        )
                    )
                return {
                    "identity": identity,
                    "permissions": checked_permissions,
                }

            # redefine GET /me because we want our current_user dependency
            # it is first defined in users_router and so it wins over the one in
            # fapi_users.get_users_router
            users_router = APIRouter()

            @users_router.get("/me")
            async def get_me(
                user: UserRead = Depends(backend.current_user(permissions={"admin": ["read"]})),
            ):
                return user

            users_router.include_router(backend.fapi_users.get_users_router(UserRead, UserUpdate))

            # Cookie based auth login and logout
            self.include_router(
                backend.fapi_users.get_auth_router(backend.cookie_authentication), prefix="/auth"
            )
            self.include_router(
                backend.fapi_users.get_register_router(UserRead, UserCreate),
                prefix="/auth",
                dependencies=[Depends(backend.current_user(permissions={"admin": ["write"]}))],
            )
            self.include_router(users_router, prefix="/auth/user")

            # GitHub OAuth register router
            self.include_router(
                backend.fapi_users.get_oauth_router(
                    backend.github_authentication, backend.github_cookie_authentication, db.secret
                ),
                prefix="/auth/github",
            )
            self.include_router(router)

            self.create_user = create_user
            self.__update_user = update_user
            self.get_user_by_email = get_user_by_email

        async def _update_user(self, user, **kwargs):
            return await self.__update_user(user, **kwargs)

        def current_user(self, permissions: Optional[Dict[str, List[str]]] = None) -> Callable:
            return backend.current_user(permissions)

        async def update_user(self, update_user=Depends(backend.update_user)) -> Callable:
            return update_user

        def websocket_auth(
            self,
            permissions: Optional[Dict[str, List[str]]] = None,
        ) -> Callable[[], Tuple[Any, Dict[str, List[str]]]]:
            return backend.websocket_auth(permissions)

    return _Auth()


# From https://en.wikipedia.org/wiki/Moons_of_Jupiter
moons_of_jupyter = (
    "Metis",
    "Adrastea",
    "Amalthea",
    "Thebe",
    "Io",
    "Europa",
    "Ganymede",
    "Callisto",
    "Themisto",
    "Leda",
    "Ersa",
    "Pandia",
    "Himalia",
    "Lysithea",
    "Elara",
    "Dia",
    "Carpo",
    "Valetudo",
    "Euporie",
    "Eupheme",
    # 'S/2003 J 18',
    # 'S/2010 J 2',
    "Helike",
    # 'S/2003 J 16',
    # 'S/2003 J 2',
    "Euanthe",
    # 'S/2017 J 7',
    "Hermippe",
    "Praxidike",
    "Thyone",
    "Thelxinoe",
    # 'S/2017 J 3',
    "Ananke",
    "Mneme",
    # 'S/2016 J 1',
    "Orthosie",
    "Harpalyke",
    "Iocaste",
    # 'S/2017 J 9',
    # 'S/2003 J 12',
    # 'S/2003 J 4',
    "Erinome",
    "Aitne",
    "Herse",
    "Taygete",
    # 'S/2017 J 2',
    # 'S/2017 J 6',
    "Eukelade",
    "Carme",
    # 'S/2003 J 19',
    "Isonoe",
    # 'S/2003 J 10',
    "Autonoe",
    "Philophrosyne",
    "Cyllene",
    "Pasithee",
    # 'S/2010 J 1',
    "Pasiphae",
    "Sponde",
    # 'S/2017 J 8',
    "Eurydome",
    # 'S/2017 J 5',
    "Kalyke",
    "Hegemone",
    "Kale",
    "Kallichore",
    # 'S/2011 J 1',
    # 'S/2017 J 1',
    "Chaldene",
    "Arche",
    "Eirene",
    "Kore",
    # 'S/2011 J 2',
    # 'S/2003 J 9',
    "Megaclite",
    "Aoede",
    # 'S/2003 J 23',
    "Callirrhoe",
    "Sinope",
)


def get_anonymous_username() -> str:
    """
    Get a random user-name based on the moons of Jupyter.
    This function returns names like "Anonymous Io" or "Anonymous Metis".
    """
    return moons_of_jupyter[random.randint(0, len(moons_of_jupyter) - 1)]



from abc import ABC, abstractmethod

from fastapi import APIRouter

from jupyverse_api import Router

from ..app import App


class Login(Router, ABC):
    def __init__(self, app: App):
        super().__init__(app)

        router = APIRouter()

        @router.get("/login")
        async def get_login():
            return await self.get_login()

        self.include_router(router)

    @abstractmethod
    async def get_login(self):
        ...



from functools import partial
from typing import Any, Callable, Optional

from pycrdt import Doc, Text

from .ybasedoc import YBaseDoc


class YUnicode(YBaseDoc):
    def __init__(self, ydoc: Optional[Doc] = None):
        super().__init__(ydoc)
        self._ysource = Text()
        self._ydoc["source"] = self._ysource

    @property
    def version(self) -> str:
        return "1.0.0"

    def get(self) -> str:
        return str(self._ysource)

    def set(self, value: str) -> None:
        with self._ydoc.transaction():
            # clear document
            del self._ysource[:]
            # initialize document
            if value:
                self._ysource += value

    def observe(self, callback: Callable[[str, Any], None]) -> None:
        self.unobserve()
        self._subscriptions[self._ystate] = self._ystate.observe(partial(callback, "state"))
        self._subscriptions[self._ysource] = self._ysource.observe(partial(callback, "source"))



import sys
from typing import Any

try:
    import ypywidgets  # noqa: F401
    from pycrdt import (
        TransactionEvent,
        YMessageType,
        YSyncMessageType,
        create_sync_message,
        create_update_message,
        handle_sync_message,
    )
    ypywidgets_installed = True
except ImportError:
    ypywidgets_installed = False

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points


Widgets: Any

if ypywidgets_installed:
    class Widgets:  # type: ignore
        def __init__(self):
            self.ydocs = {
                ep.name: ep.load() for ep in entry_points(group="ypywidgets")
            }
            self.widgets = {}

        def comm_open(self, msg, comm) -> None:
            target_name = msg["content"]["target_name"]
            if target_name != "ywidget":
                return

            name = msg["metadata"]["ymodel_name"]
            comm_id = msg["content"]["comm_id"]
            self.comm = comm
            model = self.ydocs[f"{name}Model"]()
            self.widgets[comm_id] = {"model": model, "comm": comm}
            msg = create_sync_message(model.ydoc)
            comm.send(buffers=[msg])

        def comm_msg(self, msg) -> None:
            comm_id = msg["content"]["comm_id"]
            message = bytes(msg["buffers"][0])
            if message[0] == YMessageType.SYNC:
                ydoc = self.widgets[comm_id]["model"].ydoc
                reply = handle_sync_message(
                    message[1:],
                    ydoc,
                )
                if reply:
                    self.widgets[comm_id]["comm"].send(buffers=[reply])
                if message[1] == YSyncMessageType.SYNC_STEP2:
                    ydoc.observe(self._send)

        def _send(self, event: TransactionEvent):
            update = event.update  # type: ignore
            message = create_update_message(update)
            try:
                self.comm.send(buffers=[message])
            except Exception:
                pass
else:
    Widgets = None



import sys
from copy import deepcopy
from pathlib import Path
from uuid import uuid4

import easywebdav  # type: ignore
import pytest
from anyio import to_thread
from asphalt.core import Context

from jupyverse_api.main import JupyverseComponent

COMPONENTS = {
    "app": {"type": "app"},
    "webdav": {"type": "webdav"},
}


def configure(components, config):
    _components = deepcopy(components)
    for k1, v1 in config.items():
        for k2, v2 in v1.items():
            _components[k1][k2] = v2
    return _components


@pytest.mark.asyncio
@pytest.mark.skipif(sys.version_info < (3, 10), reason="Requires Python >=3.10")
async def test_webdav(unused_tcp_port):
    components = configure(
        COMPONENTS, {"webdav": {"account_mapping": [{"username": "foo", "password": "bar"}]}}
    )
    async with Context() as ctx:
        await JupyverseComponent(
            components=components,
            port=unused_tcp_port,
        ).start(ctx)

        webdav = easywebdav.connect(
            "127.0.0.1", port=unused_tcp_port, path="webdav", username="foo", password="bar"
        )
        dirname = str(uuid4())
        assert not Path(dirname).exists()
        await to_thread.run_sync(webdav.mkdir, dirname)
        assert Path(dirname).is_dir()
        await to_thread.run_sync(webdav.rmdir, dirname)
        assert not Path(dirname).exists()



from __future__ import annotations

from contextlib import AsyncExitStack
from logging import Logger, getLogger

from anyio import TASK_STATUS_IGNORED, Event, create_task_group
from anyio.abc import TaskGroup, TaskStatus
from pycrdt import Doc

from .websocket import Websocket
from .yroom import YRoom


class WebsocketServer:
    """WebSocket server."""

    auto_clean_rooms: bool
    rooms: dict[str, YRoom]
    _started: Event | None
    _starting: bool
    _task_group: TaskGroup | None

    def __init__(
        self, rooms_ready: bool = True, auto_clean_rooms: bool = True, log: Logger | None = None
    ) -> None:
        """Initialize the object.

        The WebsocketServer instance should preferably be used as an async context manager:
        ```py
        async with websocket_server:
            ...
        ```
        However, a lower-level API can also be used:
        ```py
        task = asyncio.create_task(websocket_server.start())
        await websocket_server.started.wait()
        ...
        websocket_server.stop()
        ```

        Arguments:
            rooms_ready: Whether rooms are ready to be synchronized when opened.
            auto_clean_rooms: Whether rooms should be deleted when no client is there anymore.
            log: An optional logger.
        """
        self.rooms_ready = rooms_ready
        self.auto_clean_rooms = auto_clean_rooms
        self.log = log or getLogger(__name__)
        self.rooms = {}
        self._started = None
        self._starting = False
        self._task_group = None

    @property
    def started(self) -> Event:
        """An async event that is set when the WebSocket server has started."""
        if self._started is None:
            self._started = Event()
        return self._started

    async def get_room(self, name: str, ydoc: Doc | None = None) -> YRoom:
        """Get or create a room with the given name, and start it.

        Arguments:
            name: The room name.

        Returns:
            The room with the given name, or a new one if no room with that name was found.
        """
        if name not in self.rooms.keys():
            self.rooms[name] = YRoom(ydoc=ydoc, ready=self.rooms_ready, log=self.log)
        room = self.rooms[name]
        await self.start_room(room)
        return room

    async def start_room(self, room: YRoom) -> None:
        """Start a room, if not already started.

        Arguments:
            room: The room to start.
        """
        if self._task_group is None:
            raise RuntimeError(
                "The WebsocketServer is not running: use `async with websocket_server:` "
                "or `await websocket_server.start()`"
            )

        if not room.started.is_set():
            await self._task_group.start(room.start)

    def get_room_name(self, room: YRoom) -> str:
        """Get the name of a room.

        Arguments:
            room: The room to get the name from.

        Returns:
            The room name.
        """
        return list(self.rooms.keys())[list(self.rooms.values()).index(room)]

    def rename_room(
        self, to_name: str, *, from_name: str | None = None, from_room: YRoom | None = None
    ) -> None:
        """Rename a room.

        Arguments:
            to_name: The new name of the room.
            from_name: The previous name of the room (if `from_room` is not passed).
            from_room: The room to be renamed (if `from_name` is not passed).
        """
        if from_name is not None and from_room is not None:
            raise RuntimeError("Cannot pass from_name and from_room")
        if from_name is None:
            assert from_room is not None
            from_name = self.get_room_name(from_room)
        self.rooms[to_name] = self.rooms.pop(from_name)

    def delete_room(self, *, name: str | None = None, room: YRoom | None = None) -> None:
        """Delete a room.

        Arguments:
            name: The name of the room to delete (if `room` is not passed).
            room: The room to delete ( if `name` is not passed).
        """
        if name is not None and room is not None:
            raise RuntimeError("Cannot pass name and room")
        if name is None:
            assert room is not None
            name = self.get_room_name(room)
        room = self.rooms.pop(name)
        room.stop()

    async def serve(self, websocket: Websocket) -> None:
        """Serve a client through a WebSocket.

        Arguments:
            websocket: The WebSocket through which to serve the client.
        """
        if self._task_group is None:
            raise RuntimeError(
                "The WebsocketServer is not running: use `async with websocket_server:` "
                "or `await websocket_server.start()`"
            )

        async with create_task_group() as tg:
            tg.start_soon(self._serve, websocket, tg)

    async def _serve(self, websocket: Websocket, tg: TaskGroup):
        room = await self.get_room(websocket.path)
        await self.start_room(room)
        await room.serve(websocket)

        if self.auto_clean_rooms and not room.clients:
            self.delete_room(room=room)
        tg.cancel_scope.cancel()

    async def __aenter__(self) -> WebsocketServer:
        if self._task_group is not None:
            raise RuntimeError("WebsocketServer already running")

        async with AsyncExitStack() as exit_stack:
            tg = create_task_group()
            self._task_group = await exit_stack.enter_async_context(tg)
            self._exit_stack = exit_stack.pop_all()
            self.started.set()

        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        if self._task_group is None:
            raise RuntimeError("WebsocketServer not running")

        self._task_group.cancel_scope.cancel()
        self._task_group = None
        return await self._exit_stack.__aexit__(exc_type, exc_value, exc_tb)

    async def start(self, *, task_status: TaskStatus[None] = TASK_STATUS_IGNORED):
        """Start the WebSocket server.

        Arguments:
            task_status: The status to set when the task has started.
        """
        if self._starting:
            return
        else:
            self._starting = True

        if self._task_group is not None:
            raise RuntimeError("WebsocketServer already running")

        # create the task group and wait forever
        async with create_task_group() as self._task_group:
            self._task_group.start_soon(Event().wait)
            self.started.set()
            self._starting = False
            task_status.started()

    def stop(self) -> None:
        """Stop the WebSocket server."""
        if self._task_group is None:
            raise RuntimeError("WebsocketServer not running")

        self._task_group.cancel_scope.cancel()
        self._task_group = None



import contextlib

from fps_auth.backends import get_backend
from fps_auth.config import _AuthConfig
from fps_auth.db import get_db
from fps_auth.models import UserCreate, UserUpdate
from rich.console import Console
from rich.table import Table
from sqlalchemy import select

from jupyverse_api.frontend import FrontendConfig

auth_config = _AuthConfig()
frontend_config = FrontendConfig()
db = get_db(auth_config)
backend = get_backend(auth_config, frontend_config, db)

get_async_session_context = contextlib.asynccontextmanager(db.get_async_session)
get_user_db_context = contextlib.asynccontextmanager(db.get_user_db)
get_user_manager_context = contextlib.asynccontextmanager(backend.get_user_manager)


async def get_user_by_email(email):
    async with get_async_session_context() as session:
        async with get_user_db_context(session) as user_db:
            async with get_user_manager_context(user_db) as user_manager:
                return await user_manager.get_by_email(email)


async def delete_user(user):
    async with get_async_session_context() as session:
        async with get_user_db_context(session) as user_db:
            async with get_user_manager_context(user_db) as user_manager:
                return await user_manager.delete(user)


async def create_user(
    username: str, email: str, password: str, is_superuser: bool = False, permissions={}
):
    async with get_async_session_context() as session:
        async with get_user_db_context(session) as user_db:
            async with get_user_manager_context(user_db) as user_manager:
                return await user_manager.create(
                    UserCreate(
                        username=username,
                        email=email,
                        password=password,
                        is_superuser=is_superuser,
                        permissions=permissions,
                    )
                )


async def update_user(user_update: UserUpdate, user, safe: bool = False):
    async with get_async_session_context() as session:
        async with get_user_db_context(session) as user_db:
            async with get_user_manager_context(user_db) as user_manager:
                return await user_manager.update(user_update, user, safe)


async def get_users():
    async with get_async_session_context() as session:
        statement = select(db.User)
        users = (await session.execute(statement)).unique().all()
    return [usr.User for usr in users if usr.User.is_active]


def show_users(users, include_attrs=[], exclude_attrs=[]):
    if include_attrs:
        attrs = include_attrs
    else:
        attrs = [
            "username",
            "name",
            "email",
            "display_name",
            "permissions",
            "anonymous",
            "workspace",
            "settings",
            "avatar_url",
            "color",
            "initials",
            "oauth_accounts",
        ]
        for a in exclude_attrs:
            attrs.remove(a)
    table = Table(title="Users")
    for a in attrs:
        table.add_column(a)
    for user in users:
        table.add_row(*(str(getattr(user, a)) for a in attrs))
    console = Console()
    console.print(table)



import logging

from asphalt.core import Component, Context
from fastapi_users.exceptions import UserAlreadyExists

from jupyverse_api.app import App
from jupyverse_api.auth import Auth, AuthConfig
from jupyverse_api.frontend import FrontendConfig
from jupyverse_api.main import Host, QueryParams

from .config import _AuthConfig
from .routes import auth_factory

logger = logging.getLogger("auth")


class AuthComponent(Component):
    def __init__(self, **kwargs):
        self.auth_config = _AuthConfig(**kwargs)

    async def start(
        self,
        ctx: Context,
    ) -> None:
        ctx.add_resource(self.auth_config, types=AuthConfig)

        app = await ctx.request_resource(App)
        frontend_config = await ctx.request_resource(FrontendConfig)

        auth = auth_factory(app, self.auth_config, frontend_config)
        ctx.add_resource(auth, types=Auth)

        await auth.db.create_db_and_tables()

        if self.auth_config.test:
            try:
                await auth.create_user(
                    username="admin@jupyter.com",
                    email="admin@jupyter.com",
                    password="jupyverse",
                    permissions={"admin": ["read", "write"]},
                )
            except UserAlreadyExists:
                pass

        try:
            await auth.create_user(
                username=self.auth_config.token,
                email=self.auth_config.global_email,
                password="",
                permissions={},
            )
        except UserAlreadyExists:
            global_user = await auth.get_user_by_email(self.auth_config.global_email)
            await auth._update_user(
                global_user,
                username=self.auth_config.token,
                permissions={},
            )

        if self.auth_config.mode == "token":
            query_params = await ctx.request_resource(QueryParams)
            host = await ctx.request_resource(Host)
            query_params.d["token"] = self.auth_config.token

            logger.info("")
            logger.info("To access the server, copy and paste this URL:")
            logger.info(f"{host.url}?token={self.auth_config.token}")
            logger.info("")



from typing import Optional
from uuid import uuid4

from jupyverse_api.auth import AuthConfig


class _AuthConfig(AuthConfig):
    client_id: str = ""
    client_secret: str = ""
    redirect_uri: str = ""
    # mode: Literal["noauth", "token", "user"] = "token"
    mode: str = "token"
    token: str = uuid4().hex
    global_email: str = "guest@jupyter.com"
    cookie_secure: bool = False  # FIXME: should default to True, and set to False for tests
    clear_users: bool = False
    test: bool = False
    login_url: Optional[str] = None
    directory: Optional[str] = None



jupyverse_jupyterlab_command = " ".join(
    [
        "jupyverse",
        "--set auth.mode=noauth",
        "--set frontend.collaborative=true",
        "--set frontend.base_url={base_url}jupyverse-jupyterlab/",
        "--disable notebook",
        "--port={port}",
    ]
    + [">jupyverse_jupyterlab.log 2>&1"]
)


jupyverse_notebook_command = " ".join(
    [
        "jupyverse",
        "--set auth.mode=noauth",
        "--set frontend.collaborative=true",
        "--set frontend.base_url={base_url}jupyverse-notebook/",
        "--disable jupyterlab",
        "--port={port}",
    ]
    + [">jupyverse_notebook.log 2>&1"]
)


c.ServerProxy.servers = {
    "jupyverse-jupyterlab": {
        "command": ["/bin/bash", "-c", jupyverse_jupyterlab_command],
        "timeout": 60,
        "absolute_url": False,
    },
    "jupyverse-notebook": {
        "command": ["/bin/bash", "-c", jupyverse_notebook_command],
        "timeout": 60,
        "absolute_url": False,
    },
}

c.NotebookApp.default_url = "/jupyverse-jupyterlab"

import logging

c.NotebookApp.log_level = logging.DEBUG



import os
from pathlib import Path

import pytest
from asphalt.core import Context
from httpx import AsyncClient
from utils import clear_content_values, configure, create_content, sort_content_by_name

from jupyverse_api.main import JupyverseComponent

COMPONENTS = {
    "app": {"type": "app"},
    "auth": {"type": "auth", "test": True},
    "contents": {"type": "contents"},
    "frontend": {"type": "frontend"},
}


@pytest.mark.asyncio
@pytest.mark.parametrize("auth_mode", ("noauth",))
async def test_tree(auth_mode, tmp_path, unused_tcp_port):
    prev_dir = os.getcwd()
    os.chdir(tmp_path)
    dname = Path(".")
    expected = []
    # write some files with different size
    for size in range(7, 13):
        fname = f"file_size_{size}"
        with (dname / fname).open("w") as f:
            f.write("-" * size)
        expected.append(
            create_content(
                content=None,
                type="file",
                size=size,
                mimetype="text/plain",
                name=fname,
                path=(dname / fname).as_posix(),
                format=None,
            )
        )
    # write some directories
    for i in range(3):
        sub_dname = f"directory_{i}"
        (dname / sub_dname).mkdir()
        expected.append(
            create_content(
                content=None,
                type="directory",
                size=None,
                mimetype=None,
                name=sub_dname,
                path=(dname / sub_dname).as_posix(),
                format="json",
            )
        )
    expected = create_content(
        content=expected,
        type="directory",
        size=None,
        mimetype=None,
        name="",
        path=dname.as_posix(),
        format="json",
    )

    components = configure(COMPONENTS, {"auth": {"mode": auth_mode}})
    async with Context() as ctx, AsyncClient() as http:
        await JupyverseComponent(
            components=components,
            port=unused_tcp_port,
        ).start(ctx)

        response = await http.get(
            f"http://127.0.0.1:{unused_tcp_port}/api/contents", params={"content": 1}
        )
        actual = response.json()
        # ignore modification and creation times
        clear_content_values(actual, keys=["created", "last_modified"])
        # ensure content names are ordered the same way
        sort_content_by_name(actual)
        sort_content_by_name(expected)
        assert actual == expected
        os.chdir(prev_dir)



from asphalt.core import Component, Context

from jupyverse_api.app import App
from jupyverse_api.auth import Auth
from jupyverse_api.frontend import FrontendConfig
from jupyverse_api.lab import Lab
from jupyverse_api.notebook import Notebook

from .routes import _Notebook


class NotebookComponent(Component):
    async def start(
        self,
        ctx: Context,
    ) -> None:
        app = await ctx.request_resource(App)
        auth = await ctx.request_resource(Auth)  # type: ignore
        frontend_config = await ctx.request_resource(FrontendConfig)
        lab = await ctx.request_resource(Lab)  # type: ignore

        notebook = _Notebook(app, auth, frontend_config, lab)
        ctx.add_resource(notebook, types=Notebook)



import json
from http import HTTPStatus
from pathlib import Path

import jupyterlab as jupyterlab_module  # type: ignore
from fastapi import Response
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from starlette.requests import Request

from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User
from jupyverse_api.frontend import FrontendConfig
from jupyverse_api.jupyterlab import JupyterLab, JupyterLabConfig
from jupyverse_api.lab import Lab

from .index import INDEX_HTML


class _JupyterLab(JupyterLab):
    def __init__(
        self,
        app: App,
        jupyterlab_config: JupyterLabConfig,
        auth: Auth,
        frontend_config: FrontendConfig,
        lab: Lab,
    ) -> None:
        super().__init__(app, auth)

        self.jupyterlab_config = jupyterlab_config
        self.frontend_config = frontend_config
        self.lab = lab
        lab.redirect_after_root = "lab"

        extensions_dir = lab.prefix_dir / "share" / "jupyter" / "labextensions"
        self.federated_extensions, self.disabled_extensions = lab.get_federated_extensions(
            extensions_dir
        )
        jupyterlab_dir = Path(jupyterlab_module.__file__).parents[1]

        if jupyterlab_config.dev_mode:
            self.static_lab_dir = jupyterlab_dir / "dev_mode" / "static"
        else:
            self.static_lab_dir = lab.prefix_dir / "share" / "jupyter" / "lab" / "static"

        self.mount(
            "/static/lab",
            StaticFiles(directory=self.static_lab_dir),
            name="static",
        )

    async def get_lab(
        self,
        user: User,
    ):
        return HTMLResponse(
            self.get_index(
                "default",
                self.frontend_config.collaborative,
                self.jupyterlab_config.server_side_execution,
                self.jupyterlab_config.dev_mode,
                self.frontend_config.base_url,
            )
        )

    async def load_workspace(
        self,
        path,
    ):
        return HTMLResponse(
            self.get_index(
                "default",
                self.frontend_config.collaborative,
                self.jupyterlab_config.server_side_execution,
                self.jupyterlab_config.dev_mode,
                self.frontend_config.base_url,
            )
        )

    async def get_workspace_data(self, user: User):
        if user:
            return json.loads(user.workspace)
        return {}

    async def set_workspace(
        self,
        request: Request,
        user: User,
        user_update,
    ):
        workspace = (await request.body()).decode("utf-8")
        await user_update({"workspace": workspace})
        return Response(status_code=HTTPStatus.NO_CONTENT.value)

    async def get_workspace(
        self,
        name,
        user: User,
    ):
        return self.get_index(
            name,
            self.frontend_config.collaborative,
            self.jupyterlab_config.server_side_execution,
            self.jupyterlab_config.dev_mode,
            self.frontend_config.base_url,
        )

    def get_index(self, workspace, collaborative, server_side_execution, dev_mode, base_url="/"):
        for path in (self.static_lab_dir).glob("main.*.js"):
            main_id = path.name.split(".")[1]
            break
        vendor_id = None
        for path in (self.static_lab_dir).glob("vendors-node_modules_whatwg-fetch_fetch_js.*.js"):
            vendor_id = path.name.split(".")[1]
            break
        full_static_url = f"{base_url}static/lab"

        page_config = {
            "appName": "JupyterLab",
            "appNamespace": "lab",
            "appUrl": "/lab",
            "appVersion": jupyterlab_module.__version__,
            "baseUrl": base_url,
            "cacheFiles": False,
            "collaborative": collaborative,
            "serverSideExecution": server_side_execution,
            "devMode": dev_mode,
            "disabledExtensions": self.disabled_extensions,
            "exposeAppInBrowser": False,
            "extraLabextensionsPath": [],
            "federated_extensions": self.federated_extensions,
            "fullAppUrl": f"{base_url}lab",
            "fullLabextensionsUrl": f"{base_url}lab/extensions",
            "fullLicensesUrl": f"{base_url}lab/api/licenses",
            "fullListingsUrl": f"{base_url}lab/api/listings",
            "fullMathjaxUrl": f"{base_url}static/notebook/components/MathJax/MathJax.js",
            "fullSettingsUrl": f"{base_url}lab/api/settings",
            "fullStaticUrl": full_static_url,
            "fullThemesUrl": f"{base_url}lab/api/themes",
            "fullTranslationsApiUrl": f"{base_url}lab/api/translations",
            "fullTreeUrl": f"{base_url}lab/tree",
            "fullWorkspacesApiUrl": f"{base_url}lab/api/workspaces",
            "ignorePlugins": [],
            "labextensionsUrl": "/lab/extensions",
            "licensesUrl": "/lab/api/licenses",
            "listingsUrl": "/lab/api/listings",
            "mathjaxConfig": "TeX-AMS-MML_HTMLorMML-full,Safe",
            "mode": "multiple-document",
            "notebookVersion": "[1, 9, 0]",
            "quitButton": True,
            "settingsUrl": "/lab/api/settings",
            "store_id": 0,
            "schemasDir": (
                self.lab.prefix_dir / "share" / "jupyter" / "lab" / "schemas"
            ).as_posix(),
            "terminalsAvailable": True,
            "themesDir": (self.lab.prefix_dir / "share" / "jupyter" / "lab" / "themes").as_posix(),
            "themesUrl": "/lab/api/themes",
            "token": "4e2804532de366abc81e32ab0c6bf68a73716fafbdbb2098",
            "translationsApiUrl": "/lab/api/translations",
            "treePath": "",
            "workspace": workspace,
            "treeUrl": "/lab/tree",
            "workspacesApiUrl": "/lab/api/workspaces",
            "wsUrl": "",
        }
        index = (
            INDEX_HTML.replace("PAGE_CONFIG", json.dumps(page_config))
            .replace("FULL_STATIC_URL", full_static_url)
            .replace("MAIN_ID", main_id)
        )
        if vendor_id:
            index = index.replace(
                "VENDORS_NODE_MODULES",
                '<script defer src="/static/lab/vendors-node_modules_whatwg-fetch_fetch_js.'
                f'{vendor_id}.js"></script>',
            )
        else:
            index = index.replace("VENDORS_NODE_MODULES", "")
        return index



from __future__ import annotations

import asyncio
from collections.abc import AsyncGenerator
from pathlib import Path
from typing import Optional

from asphalt.core import Component, Context, context_teardown

from jupyverse_api.app import App
from jupyverse_api.auth import Auth
from jupyverse_api.frontend import FrontendConfig
from jupyverse_api.kernels import Kernels, KernelsConfig
from jupyverse_api.yjs import Yjs

from .kernel_driver.paths import jupyter_runtime_dir
from .routes import _Kernels


class KernelsComponent(Component):
    def __init__(self, **kwargs):
        self.kernels_config = KernelsConfig(**kwargs)

    @context_teardown
    async def start(
        self,
        ctx: Context,
    ) -> AsyncGenerator[None, Optional[BaseException]]:
        ctx.add_resource(self.kernels_config, types=KernelsConfig)

        app = await ctx.request_resource(App)
        auth = await ctx.request_resource(Auth)  # type: ignore
        frontend_config = await ctx.request_resource(FrontendConfig)
        yjs = (
            await ctx.request_resource(Yjs)  # type: ignore
            if self.kernels_config.require_yjs
            else None
        )

        kernels = _Kernels(app, self.kernels_config, auth, frontend_config, yjs)
        ctx.add_resource(kernels, types=Kernels)

        if self.kernels_config.allow_external_kernels:
            external_connection_dir = self.kernels_config.external_connection_dir
            if external_connection_dir is None:
                path = Path(jupyter_runtime_dir()) / "external_kernels"
            else:
                path = Path(external_connection_dir)
            task = asyncio.create_task(kernels.watch_connection_files(path))

        yield

        if self.kernels_config.allow_external_kernels:
            task.cancel()
        for kernel in kernels.kernels.values():
            await kernel["server"].stop()



from abc import ABC, abstractmethod

from fastapi import APIRouter, Depends

from jupyverse_api import Config, Router

from ..app import App
from ..auth import Auth, User


class ResourceUsage(Router, ABC):
    def __init__(self, app: App, auth: Auth):
        super().__init__(app)

        router = APIRouter()

        @router.get("/api/metrics/v1")
        async def get_metrics(
            user: User = Depends(auth.current_user(permissions={"contents": ["read"]})),
        ):
            return await self.get_metrics(user)

        self.include_router(router)

    @abstractmethod
    async def get_metrics(
        self,
        user: User,
    ):
        ...


class ResourceUsageConfig(Config):
    mem_limit: int = 0
    mem_warning_threshold: int = 0
    track_cpu_percent: bool = False
    cpu_limit: int = 0
    cpu_warning_threshold: int = 0
